from gerel.genome.factories import minimal, copy
from gerel.mutators.mutator import Mutator


def curry_genome_seeder(mutator, seed_genomes=None):
    """Construct genome generator.

    Given a mutator and a collection of initial genomes builds a
    function that can then return as many as needed. Used for
    populating Population Objects.

    :param mutator: defined Mutator object that acts on genomes.
    :param seed_genomes: list of genomes that will be copied and mutated.
    :return: Generator function that takes a value n and returns n genomes
      generated by sampling the seed_genome list, copying and mutating.
    :raises ValueError: mutator param must be of type Mutator.
    """
    if not seed_genomes:
        seed_genomes = [minimal()]

    if not issubclass(type(mutator), Mutator):
        raise ValueError('Mutator param must be of type Mutator')

    def genome_seeder(n):
        count = 0
        while count < n:
            for seed_genome in seed_genomes:
                genome = copy(seed_genome)
                mutator(genome)
                yield genome
                count += 1
    return genome_seeder



import itertools


class Edge:
    """Edge class.

    Represents a directed edge on a genome (graph structure). Stores
    reference to from_node and to_node. A weight of the edge and
    also it's innovation number.
    """

    innov_iter = itertools.count()
    registry = {}

    def __init__(self, from_node, to_node, weight, active=True):
        """Creates edge between from_node and to_node with
        specified weight and activity.

        :param from_node: Source Node in genome
        :param to_node: Target Node in genome
        :param weight: Weight of edge, represents contribution of
            from_nodes input to to_node
        :param active: If edge not active it will not effect a
            model class built with to_reduced_repr object.

        :raises ValueError: to_node must be a member of a layer
            further on in the genome.
        """
        innov = Edge.registry.get((from_node.innov, to_node.innov), None)
        innov = innov if innov is not None else next(Edge.innov_iter)
        Edge.registry[(from_node.innov, to_node.innov)] = innov
        self.innov = innov
        if to_node.layer_num - from_node.layer_num < 1:
            raise ValueError('Cannot connect edge to lower or same layer')
        self.active = active
        self.from_node = from_node
        self.to_node = to_node
        self.weight = weight
        from_node.edges_out.append(self)
        to_node.edges_in.append(self)

    @classmethod
    def copy(cls,
             edge,
             new_genome):
        """Copies a edge onto a genome.

        :param edge: Edge to copy and place on new_genome
        :param new_genome: Target of copied edge.
        :return: The copied edge.
        :raises ValueError: from and to nodes must exist in the target
            genome. The added edges innovation number must be in order
            with current edge innovation numbers.
        """
        fn = edge.from_node
        tn = edge.to_node

        if len(new_genome.layers) - 1 < fn.layer_num or \
                len(new_genome.layers[fn.layer_num]) - 1 < fn.layer_ind:
            raise ValueError('from_node does not exist on new_genome.')

        if len(new_genome.layers) - 1 < tn.layer_num or \
                len(new_genome.layers[tn.layer_num]) - 1 < tn.layer_ind:
            raise ValueError('to_node does not exist on new_genome.')

        from_node = new_genome.layers[fn.layer_num][fn.layer_ind]
        to_node = new_genome.layers[tn.layer_num][tn.layer_ind]

        edge = cls(from_node, to_node, edge.weight)
        if new_genome.edges and new_genome.edges[-1].innov > edge.innov:
            raise ValueError('innovation numbers are out of order.')
        new_genome.edges.append(edge)
        return edge

    @property
    def to_reduced_repr(self):
        """Creates simple serializable representation.

        :return: tuple of index and property values.
        """
        return (
            self.from_node.to_reduced_repr,
            self.to_node.to_reduced_repr,
            self.weight,
            self.innov,
            self.active
        )



import json
import numpy as np
import os
from datetime import datetime


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)


class DataStore:
    def __init__(
            self,
            dirname=os.getcwd(),
            name='data',
            params=None):

        params = params if params else {}
        self.params = {
            'datetime': str(datetime.now()),
            **params
        }
        self.name = name
        self.dirname = os.path.join(dirname, self.name)
        if not os.path.exists(self.dirname):
            os.makedirs(self.dirname)

    def save(self, data):
        fname = os.path.join(self.dirname, str(len(os.listdir(self.dirname)) + 1))
        with open(fname, 'w') as file:
            file.write(json.dumps(data, cls=NpEncoder))

    def load(self, index):
        fname = os.path.join(self.dirname, str(index))
        with open(fname, 'r') as file:
            return json.loads(file.read())

    def generations(self):
        for file_name in sorted(os.listdir(self.dirname), key=lambda fn_str: int(fn_str)):
            file_name = os.path.join(self.dirname, file_name)
            with open(file_name, 'r') as file:
                yield json.loads(file.read())

    @property
    def num_generations(self):
        return len(os.listdir(self.dirname))

    @property
    def last_generation_id(self):
        return max(os.listdir(self.dirname))

    @property
    def best_generation(self):
        best_fitness = None
        best_generation = 0
        for gen_id in os.listdir(self.dirname):
            data = self.load(gen_id)
            if not best_fitness or data['best_fitness'] > best_fitness:
                best_fitness = data['best_fitness']
                best_generation = data
        return best_generation



from gerel.genome.factories import copy
from gerel.genome.genome import Genome
from gerel.populations.population import Population
from gerel.mutators.mutator import Mutator
from numpy.random import normal
import numpy as np


class SIMPLEMutator(Mutator):
    def __init__(self, std_dev, survival_rate):
        """ SIMPLE-ES algorithm mutator.

        Orders the population by fitness and retains the top by
        specified survival_rate.

        :param std_dev: The standard deviation of the normal distribution.
        :param survival_rate: The proportion of population members that survive.
        """
        super().__init__()
        self.std_dev = np.array(std_dev)
        self.survival_rate = survival_rate

    def __call__(self, target):
        if isinstance(target, Genome):
            self.call_on_genome(target)
        elif isinstance(target, Population):
            self.call_on_population(target)

    def call_on_population(self, population):
        """Mutate Population.

        sorts genomes by fitness, removes any past a certain cutoff.
        Randomly samples from the remaining genomes and mutates them
        until the population is compelte again.

        :param population: Population Object being mutated
        :return: None
        """
        genomes = sorted(population.genomes, key=lambda g: g.fitness, reverse=True)
        cutoff = int(self.survival_rate * population.population_size)
        genomes = genomes[0: cutoff]
        population.genomes = [*genomes]
        new_genomes = np.random.choice(genomes, population.population_size - cutoff)
        for genome in new_genomes:
            new_genome = copy(genome)
            self.call_on_genome(new_genome)
            population.genomes.append(new_genome)
        population.generation += 1

    def call_on_genome(self, genome):
        """Mutate Genome.

        perturbs genomes weights by a value sampled from a normal
        distribution.

        :param genome: Genome to mutate.
        :return: None
        """
        new_weights = np.random.normal(loc=genome.weights, scale=self.std_dev)
        genome.weights = new_weights



import numpy as np
from numpy.random import choice
from gerel.genome.factories import copy
from random import random
from gerel.mutators.mutator import Mutator
from gerel.algorithms.NEAT.functions import curry_weight_mutator, curry_crossover, add_node, add_edge


class NEATMutator(Mutator):
    def __init__(
            self,
            weight_mutation_likelihood=0.8,
            weight_mutation_rate_random=0.1,
            weight_mutation_rate_uniform=0.9,
            weight_mutation_variance=0.1,
            mutation_without_crossover_rate=0.25,
            interspecies_mating_rate=0.001,
            species_member_survival_rate=0.2,
            gene_disable_rate=0.75,
            new_node_probability=0.03,
            new_edge_probability=0.05,
            weight_low=-2,
            weight_high=2,
            ):
        """build NEATMutator object that acts on NEATPopulation objects.

        :param weight_mutation_likelihood: Given a node or edge this
            is the likelihood that targets weight is mutated.
        :param weight_mutation_rate_random: Likelihood of normal
            distribution perturbation of weight.
        :param weight_mutation_variance: Variance of normal distribution
            used to perturb weights.
        :param mutation_without_crossover_rate: Likelihood of just
            weight or topological mutation occurring without crossover.
        :param interspecies_mating_rate: Likelihood of crossover between
            two species.
        :param species_member_survival_rate: Proportion of species
            members that survive yeah generation.
        :param gene_disable_rate: Likelihood of disabled gene staying
            inactive.
        :param new_node_probability: Likelihood of new node mutation.
        :param new_edge_probability: Likelihood of new edge mutation.
        :param weight_low: uniform distribution lower bound
        :param weight_high: uniform distribution upper bound
        """
        super().__init__()
        self.gene_disable_rate = gene_disable_rate
        self.new_node_probability = new_node_probability
        self.new_edge_probability = new_edge_probability
        self.mutation_without_crossover_rate = mutation_without_crossover_rate
        self.interspecies_mating_rate = interspecies_mating_rate
        self.species_member_survival_rate = species_member_survival_rate

        self.weight_mutator = curry_weight_mutator(
            weight_mutation_likelihood,
            weight_mutation_rate_random,
            weight_mutation_variance,
            weight_mutation_rate_uniform,
            weight_low=weight_low,
            weight_high=weight_high,
        )

        self.crossover = curry_crossover(gene_disable_rate)

    def call_on_population(self, population):
        """Takes population of speciated genomes and evolves them into the next generation of genomes.

        - First we compute the population proportion that each group is granted.
        - Then we keep only the top species_member_survival_rate of each generation.
        - for each group
            - we put the top performing genome into the new populations
            - randomly draw Genomes from the remaining top performing
                genomes and apply mutations/pairing until the rest of the
                groups population share is taken up.

        :param population: NEATPopulation object
        :return: None
        """
        total_group_fitness_sum = sum([item['group_fitness'] for key, item in population.species.items()])
        new_genomes = []
        for key, item in population.species.items():
            pop_prop = int(round(population.population_size * (item['group_fitness'] / total_group_fitness_sum)))
            survival_prop = int(len(item['group']) * self.species_member_survival_rate)
            survival_prop = 5 if survival_prop < 5 else survival_prop
            item['group'] = item['group'][:survival_prop]
            best_performer = copy(item['group'][0])
            new_genomes.append(best_performer)
            for _ in range(pop_prop - 1):
                selected_gene = choice(item['group'])
                new_genome = self.call_on_genome(selected_gene)
                if random() > self.mutation_without_crossover_rate:
                    other_genome = None
                    if random() < self.interspecies_mating_rate and len(population.species) > 1:
                        # select from other species
                        other_item = choice([item for _, item in population.species.items()])
                        if len(other_item['group']) > 2:
                            other_genome = choice([g for g in other_item['group'] if g is not selected_gene])
                    elif len(item['group']) > 2:
                        other_genome = choice([g for g in item['group'] if g is not selected_gene])
                    if other_genome:
                        secondary, primary = sorted([new_genome, other_genome], key=lambda g: g.fitness)
                        new_genome = self.crossover(primary, secondary)
                new_genomes.append(new_genome)
        population.generation += 1
        population.genomes = new_genomes

    def call_on_genome(self, genome):
        """Action on genome.

        Only performs weight and topological mutations.

        :param genome: Genome to copy and mutate.
        :return: New genome.
        """
        new_genome = copy(genome)
        new_genome.fitness = genome.fitness

        for edge in new_genome.edges:
            self.weight_mutator(edge)
        for node in new_genome.nodes:
            self.weight_mutator(node)

        if np.random.uniform(0, 1, 1) < self.new_node_probability:
            add_node(new_genome)

        if np.random.uniform(0, 1, 1) < self.new_edge_probability:
            add_edge(new_genome)

        return new_genome



"""Model.py

Transforms Genome object into a graph structure that can
run computation.
"""
import numpy as np
from gerel.util.activations import step
from gerel.debug.class_debug_decorator import add_inst_validator
from gerel.debug.model_validator import validate_model


@add_inst_validator(env='TESTING', validator=validate_model)
class Model:
    """Model class that acts as the expression of a genome.

    Genomes by themselves are inert but can be transformed
    into models using Model(genome.to_reduced_repr). We use
    this pattern becuase genome.to_reduced_repr is
    serializable and so can be easily passed to different
    processes or across networks.
    """

    def __init__(self, genes, activation=step, layer_activations=None):
        """ Build Model Instance.

        :param genes: Reduced representation of a genome.
            Looks like:
            (
                [(layer_num, layer_ind, innov, weight, type), ...],
                [(
                    from_node.to_reduced_repr,
                    to_node.to_reduced_repr,
                    weight,
                    innov,
                    active
                ), ...]
            )
        """
        nodes, edges = genes
        num_layer = max([layer for layer, _, _, _, _ in nodes])
        self.cells = {}
        self.layers = []

        if layer_activations and len(layer_activations) != num_layer:
            raise ValueError('Mismatch in number of activation functions and layers')

        if not layer_activations:
            # first activation function defaults to identity mapping
            layer_activations = [lambda x: x, *[activation for layer in range(num_layer - 1)]]

        for from_layer in range(num_layer):
            from_nodes = [(node_layer, node_ind, bias) for node_layer, node_ind, _, bias, _
                          in nodes if from_layer == node_layer]
            activate_weight = lambda w, a: w if a else 0
            from_edges = [(from_node_layer,
                           from_node_layer_ind,
                           to_node_layer,
                           to_node_layer_ind,
                           activate_weight(weight, active))
                for (from_node_layer, from_node_layer_ind, _, _, _),
                    (to_node_layer, to_node_layer_ind, _, _, _),
                    weight, _, active
                in edges if from_node_layer == from_layer]

            to_nodes = list(set((to_node_layer, to_node_layer_ind, bias) for
                                (from_node_layer, _, _, _, _),
                                (to_node_layer, to_node_layer_ind, _, bias, _),
                                weight, _, active in edges if from_node_layer == from_layer))

            for i, j, b in [*from_nodes, *to_nodes]:
                cell = self.cells.get((i, j), None)
                if not cell:
                    cell = Cell(i, j, b)
                    self.cells[(i, j)] = cell

            dims = (len(from_nodes), len(to_nodes))
            layer = Layer(
                dims,
                self.cells,
                from_edges,
                activation=layer_activations[from_layer])
            self.layers.append(layer)

        self.inputs = [self.cells[(i, j)] for i, j, _, _, type in nodes if type == 'input']
        self.outputs = [self.cells[(i, j)] for i, j, _, _, type in nodes if type == 'output']

    def __call__(self, inputs):
        """ Computes model output given input.

        Applies matrix multiplication to each layer states, offsets
        the next layer inputs by there node biases and then activates
        and the nodes according to the activation function:

            1/(1+exp(-c*x))

        :param inputs: list of float values.
        :return: list of float values
        """
        for cell, val in zip(self.inputs, inputs):
            cell.acc = val

        for layer in self.layers:
            layer.run()

        output = [cell.acc + cell.b for cell in self.outputs]
        self.reset()
        return output

    def reset(self):
        """Clears accumulated values after network call.

        :return: None
        """
        for _, cell in self.cells.items():
            cell.acc = 0


class Layer:
    """Layer class.

    Models are made up of layers are parts of the overall Model
    function. They first take input and compute activate signals
    for each node (Here called cell) and then perform the matrix
    multiplication operation associated to the edge weights.
    Thus the output of the layer is the next layers input.

    Note that edge connections aren't nesseserily just between
    adjacent layers in the model. An edge might skip layers so
    when the layer performs its computation it changes all cells
    connected to this layer by an edge.
    """
    def __init__(self, dims, cells, edges, activation=step):
        """ Build Layer object.

        :param dims: length 2 tuple giveing the number of layer
            inputs and outputs.
        :param cells: all cells derived from Nodes that lie in this
            genome layer.
        :param edges: all edges connected with a from_node in this
            layer.
        """
        self.activation = activation
        self.edges = edges
        self.mat = np.zeros(dims)
        input_dim, output_dim = dims
        self.inputs = []
        self.outputs = []
        input_cells_map = {}
        output_cells_map = {}
        for fn_i, fn_j, tn_i, tn_j, w in edges:
            input_cells_map[(fn_i, fn_j)] = input_cells_map.get((fn_i, fn_j), len(self.inputs))
            output_cells_map[(tn_i, tn_j)] = output_cells_map.get((tn_i, tn_j), len(self.outputs))
            if len(self.inputs) == input_cells_map[(fn_i, fn_j)]:
                self.inputs.append(cells[(fn_i, fn_j)])
            if len(self.outputs) == output_cells_map[(tn_i, tn_j)]:
                self.outputs.append(cells[(tn_i, tn_j)])
            self.mat[input_cells_map[(fn_i, fn_j)], output_cells_map[(tn_i, tn_j)]] = w

    def run(self):
        """Performs layer computation and updates relevent Model cells.

        :param activation: function taking float and returning float.
        :return: None
        """
        input_vals = np.array([self.activation(cell.acc + cell.b) for cell in self.inputs])
        output_vals = self.mat.T @ input_vals
        for val, cell in zip(output_vals, self.outputs):
            cell.acc += val


class Cell:
    """Cells Object.

    Used only to store accumulated values of layer outputs as each layer
    is run.
    """
    def __init__(self, i, j, b):
        """ Build Cell object.

        :param i: layer_num
        :param j: layer_ind
        :param b: bias weight
        """
        self.i = i
        self.j = j
        self.b = b
        self.acc = 0



from gerel.genome.genome import Genome
from gerel.populations.population import Population
from gerel.mutators.mutator import Mutator
from numpy.random import normal
import numpy as np
from gerel.util.math import vector_decomp


class RESMutator(Mutator):
    def __init__(self, initial_mu, std_dev, alpha=0.0001):
        """ REINFORCE-ES algorithm mutator.

        see:
            - https://openai.com/blog/evolution-strategies/
            - https://blog.otoro.net/2017/10/29/visual-evolution-strategies/
            - https://www.jmlr.org/papers/volume15/wierstra14a/wierstra14a.pdf

        :param initial_mu: The initial mean for the normal distribution from which
            we sample new genomes.
        :param std_dev: The standard deviation of the normal distribution.
        :param alpha: The size of update

        :raises ValueError: If initial_mu is not a non-empty list or mu and std_dev not the
            same size
        """
        super().__init__()
        if not isinstance(initial_mu, list) and not isinstance(initial_mu, np.ndarray):
            raise ValueError('initial_mu must be a list or numpy array')
        if len(initial_mu) == 0:
            raise ValueError('initial_mu must have length greater than 0')
        if isinstance(std_dev, list) and len(initial_mu) != len(std_dev):
            raise ValueError('std_dev must be a float value or a list of the same size as mu.')

        self.mu = np.array(initial_mu)
        self.std_dev = np.array(std_dev)
        self.alpha = alpha

    def __call__(self, target):
        if isinstance(target, Genome):
            self.call_on_genome(target)
        elif isinstance(target, Population):
            self.call_on_population(target)

    def compute_derivative(self, targets):
        """Computes the derivative estimate from the population.

        :param targets: list of genomes in population.
        :return: list or numpy array corresponding to gradient
        """
        grad_pi = np.zeros_like(self.mu, dtype='float64')
        for target in targets:
            f, z = target.values()
            grad_pi += f*(np.array(z) - self.mu)/(self.std_dev**2)
        return grad_pi*self.alpha/len(targets)

    def _apply_derivative(self, diff):
        self.mu = [d+m for d, m in zip(diff, self.mu)]

    def call_on_population(self, population, apply_rank_transform=True):
        """

        :param population: RESPopulation being mutated
        :param apply_rank_transform: Rank transform (normalization step)
            before mutating.
        :return: None
        """
        if apply_rank_transform:
            population.rank_transform()
        derivative = self.compute_derivative(population.genomes)
        self._apply_derivative(derivative)
        for genome in population.genomes:
            self.call_on_genome(genome)
        population.generation += 1

    def call_on_genome(self, genome):
        """Mutate genome.

        Samples a new set of genome weights from normal distribution
        centered at mu.

        :param genome: Genome being mutated.
        :return: None
        """
        new_weights = np.random.normal(loc=self.mu, scale=self.std_dev)
        genome.weights = new_weights


class ADRESMutator(RESMutator):
    def __init__(self, initial_mu, std_dev):
        """ADRESMutator object.

        **Experimental**

        Same as RESMutator but instead of updating in direction of gradient
        instead selects the best performing genome in gradient direction.

        :param initial_mu: The initial mean for the normal distribution from which
            we sample new genomes.
        :param std_dev: The standard deviation of the normal distribution.
        """
        super().__init__(initial_mu, std_dev)

    def compute_genome_preference(self, genome, derivative):
        fitness, weights = genome.values()
        weights, derivative = (np.array(weights), np.array(derivative))
        w_parallel, w_orthogonal = vector_decomp(weights, self.mu, derivative)
        if w_parallel > 0:
            preference = ((1 - w_orthogonal) ** 2) * fitness
            return preference if preference > 0 else 0
        else:
            return 0

    def update_mu(self, population, derivative):
        preference = lambda g: self.compute_genome_preference(g, derivative)
        best_genome = max(population.genomes, key=preference)
        _, new_mu = best_genome.values()
        self.mu = new_mu

    def call_on_population(self, population, apply_rank_transform=True):
        if apply_rank_transform:
            population.rank_transform()
        derivative = self.compute_derivative(population.genomes)
        normed_derivative = derivative/np.linalg.norm(derivative)
        self.update_mu(population, normed_derivative)
        for genome in population.genomes:
            self.call_on_genome(genome)
        population.generation += 1



"""Genome class."""
from itertools import chain
from gerel.genome.edge import Edge
from gerel.genome.node import Node
from gerel.util.util import sample_weight
from gerel.debug.class_debug_decorator import add_inst_validator
from gerel.debug.genome_validator import validate_genome


class DimensionMismatchError(Exception):
    """Thrown when arrays of different sizes are member wise operated on."""
    def __init__(self, message):
        self.message = message


@add_inst_validator(env="TESTING", validator=validate_genome)
class Genome:
    """Genome graph data structure.

    Encodes the data needed to generate a potential evolutionary solution to
    a given problem. Note Genomes are inert themselves and must be
    transformed into a model in order to use."""
    def __init__(
            self,
            input_size=2,
            output_size=2,
            weight_low=-2,
            weight_high=2,
            depth=3):
        """
        Builds empty genome.

        :param input_size: Number of input nodes
        :param output_size: Number of output nodes
        :param weight_low: Maximum weight on node and edges
        :param weight_high: Minimum weight on node and edges
        :param depth: Number of layers in network.
        """

        self.depth = depth
        self.weight_low = weight_low
        self.weight_high = weight_high
        self.edge_innovs = set()

        self.inputs = [Node(0, i, 0, type='input')
                       for i in range(input_size)]
        self.outputs = [Node(1 + depth, j, 0, type='output')
                        for j in range(output_size)]
        self.layers = None
        self.nodes = []
        self.edges = []
        self.fitness = None

    def layer_edges_out(self, layer_num):
        return [edge for node in self.layers[layer_num]
                for edge in node.edges_out]

    def layer_edges_in(self, layer_num):
        return [edge for node in self.layers[layer_num]
                for edge in node.edges_in]

    @property
    def weights(self):
        """All node biases and edge weights.

        :return: List of node and edge weights.
        """
        return [item.weight for item in chain(self.nodes, self.edges)]

    @weights.setter
    def weights(self, ws):
        """weights setter.

        :param ws: List of node and edge weights. Note that the weights
            must be ordered nodes and then edges.
        :raises DimensionMismatchError: if ws list is longer than genome
            nodes and edges.
        """
        if len(ws) != len(self.nodes) + len(self.edges):
            err_msg = f'Error attempting to update {len(self.nodes)} node weights and {len(self.edges)} edge weights'
            err_msg += f' with {len(ws)} update weights. Dimensions do not match.'
            raise DimensionMismatchError(err_msg)
        for weight, target in zip(ws, chain(self.nodes, self.edges)):
            target.weight = weight

    def values(self):
        return self.fitness, self.weights

    def get_admissible_edges(self):
        """Get all Admissible edges.

        Admissible edges are defined as those that span more than one
        layer and are not disabled.

        :return: List of Edges
        """

        is_admissible = lambda e: e.to_node.layer_num - e.from_node.layer_num \
            > 1 and e.active
        return [edge for layer in self.layers
                for node in layer
                for edge in node.edges_out
                if is_admissible(edge)]

    def add_node(self, layer_num):
        """Add node to genome layer.

        :param layer_num: layer to add node too. Must be less than depth.
        :raises ValueError: If attempt to add node to input or output layers.
        :return: The new node.
        """
        if layer_num == 0 or layer_num == len(self.layers) - 1:
            raise ValueError('Cannot add node to input or output layer')
        new_node = Node(layer_num, len(self.layers[layer_num]),
                        sample_weight(self.weight_low, self.weight_high))
        self.layers[layer_num].append(new_node)
        if self.nodes and new_node.innov > self.nodes[-1].innov:
            self.nodes.append(new_node)
        else:
            self.nodes = [n for n in self.nodes if n.innov < new_node.innov] + [new_node] \
                + [n for n  in self.nodes if n.innov > new_node.innov]
        return new_node

    def add_edge(self, from_node, to_node):
        """Add edge between from node and to node.

        :param from_node: Source node
        :param to_node: Target node
        :return: The new added edge.
        """
        if (from_node.innov, to_node.innov) in self.edge_innovs:
            return
        edge = Edge(
            from_node,
            to_node,
            sample_weight(self.weight_low, self.weight_high))
        # NOTE: An error was occurring here because occasionally a from and to node pair are already an edge but also
        # not currently members of the genome in question. In which case that edge is drawn from registry and inserted
        # in the next line. This causes an error because that edge is likely to be a lower innov number than the one
        # before it. Is there a better way?

        if self.edges and edge.innov > self.edges[-1].innov:
            self.edges.append(edge)
        else:
            self.edges = [e for e in self.edges if e.innov < edge.innov] + [edge] \
                + [e for e in self.edges if e.innov > edge.innov]
        self.edge_innovs.add((from_node.innov, to_node.innov))
        return edge

    def __repr__(self):
        return f'Genome(edges:{len(self.edges)}, nodes:{len(self.nodes)}, fitness:{self.fitness})'

    @property
    def to_reduced_repr(self):
        """Returns compact serializable data structure representing Genome.

        :return: List of genome node and edge values and indices.
        """
        all_nodes = [*self.inputs, *self.outputs, *self.nodes]
        return [node.to_reduced_repr for node in all_nodes], \
            [edge.to_reduced_repr for edge in self.edges]



import itertools


class Node:
    """Node Object.

    Store weight used as bias. Holds indecies of it's location in
    a genome.
    """

    innov_iter = itertools.count()
    registry = {}

    def __init__(self, layer_num, layer_ind, weight, type='hidden'):
        """Create new Node.

        :param layer_num: genome layer.
        :param layer_ind: Location in genome layer
        :param weight: Bias weight.
        :param type: 'input', 'output' or 'hidden'
        """
        innov = Node.registry.get((layer_num, layer_ind), None)
        innov = innov if innov is not None else next(Node.innov_iter)
        Node.registry[(layer_num, layer_ind)] = innov
        self.innov = innov
        self.layer_num = layer_num
        self.layer_ind = layer_ind
        self.weight = weight
        self.edges_out = []
        self.edges_in = []
        self.type = type


    @classmethod
    def copy(cls, node):
        """Deep copy of Node.

        Note the edges_out and edges_in are not copied. This is taken care of
        in the Edge initialization step itself.

        :param node: Node to be copied.
        :return: Copied Node.
        """
        return cls(
            node.layer_num,
            node.layer_ind,
            node.weight)

    @property
    def bias(self):
        return self.weight

    @property
    def to_reduced_repr(self):
        """Returns compact serializable data structure representing Node.

        :return: Tuple like: (layer_num, layer_ind, innov, weight, type)
        """
        return self.layer_num, self.layer_ind, self.innov, self.weight, self.type



def validate_population(population, *args, **kwargs):
    for key, item in population.species.items():
        if len(set(item['group'])) != len(item['group']):
            raise ValueError('Non unique genomes elements in species group.')



import random


def sample_weight(low, high):
    weight_range = high - low
    return random.random() * weight_range + low


def get_random():
    return random.random() * 2 - 1


def equal_genome(g1, g2):
    same_num_edges = len(g1.edges) == len(g2.edges)
    same_num_nodes = len(g1.nodes) == len(g2.nodes)
    same_edges = True
    for g1_edge, g2_edge in zip(g1.edges, g2.edges):
        same_edges = same_edges and g1_edge.weight == g2_edge.weight
        same_edges = same_edges and g1_edge.from_node.layer_num == g2_edge.from_node.layer_num
        same_edges = same_edges and g1_edge.from_node.layer_ind == g2_edge.from_node.layer_ind
        same_edges = same_edges and g1_edge.to_node.layer_num == g2_edge.to_node.layer_num
        same_edges = same_edges and g1_edge.to_node.layer_ind == g2_edge.to_node.layer_ind
        same_edges = same_edges and g1_edge.innov == g2_edge.innov
    same_nodes = True
    for g1_node, g2_node in zip(g1.nodes, g2.nodes):
        same_nodes = same_nodes and g1_node.weight == g2_node.weight
        same_nodes = same_nodes and g1_node.layer_num == g2_node.layer_num
        same_nodes = same_nodes and g1_node.layer_num == g2_node.layer_num
        same_nodes = same_nodes and g1_node.innov == g2_node.innov
    return same_num_nodes and same_num_edges and same_nodes and same_edges


def print_genome(genome):
    edge_innov = set(e.innov for e in genome.edges)
    node_innov = set(n.innov for n in genome.nodes)
    print('')
    print(genome)
    msg = "" if len(node_innov) == len(genome.nodes) else "(ERROR)"
    nodes, edges = genome.to_reduced_repr
    print(f' -------- nodes {msg} -------- ')
    for (a, b, i, w, type) in nodes:
        print('type: ', type, (a, b), "innov:", i, "weight:", round(w, 2))
    msg = "" if len(edge_innov) == len(genome.edges) else "(ERROR, inconsistent innov nums)"
    truth = True
    for e_1, e_2 in zip(genome.edges, genome.edges[1:]):
        truth = truth and e_1.innov < e_2.innov
    msg = msg + "" if truth else msg + "(ERROR, wrong innov order)"
    print(f' -------- edges {msg} -------- ')
    for (_, _, i_1, _, _), (_, _, i_2, _, _), w, i, active in edges:
        print(i_1, "->", i_2, "innov:", i, "active:", active, "weight:", round(w, 2))
    truth = True
    for e in genome.edges:
        truth = truth and (e.from_node.innov, e.to_node.innov) in genome.edge_innovs
    msg = "" if truth else "(ERROR)"
    print(f' -------- edge_innovs {msg} -------- ')
    print('edge_innovs:', genome.edge_innovs)


def print_population(population):
    for k, v in population.species.items():
        print('')
        print(' --- species ', k, ' --- ')
        print('representative: ', v['repr'])
        print('size: ', len(v['group']))




def validate_model(model, *args, **kwargs):
    if not model.inputs:
        raise ValueError('Model has no input nodes')

    if not model.outputs:
        raise ValueError('Model has no output nodes')



from time import time
from gerel.debug.class_debug_decorator import add_inst_validator
from gerel.debug.population_validator import validate_population
import numpy as np
from inspect import isgeneratorfunction


@add_inst_validator(env="TESTING", validator=validate_population)
class Population:
    def __init__(
            self,
            genome_seeder,
            population_size,
            delta=None,
            metric=None,
    ):
        """Construct Population of genomes.

        Stores a collection of genomes and provides methods for mutating them.

        :param genome_seeder: An iterable of genomes that will be used to generate the
            initial population.
        :param population_size: An integer number giving the size of the total Population.
        :param delta: A float number that defines the minimum distance two genomes require
            to be apart before there sorted into different species. Set to None if
            single species is desired (no speciation).
        :param metric: Function that takes two genomes and returns the distance between them.
            The metric is used to partition the genomes into species. If None then no
            speciation will take place.
        :raises ValueError: If delta is not equal to None but metric is.
        """
        if delta and not metric:
            raise ValueError('Speciation requires a metric')

        self.population_size = population_size
        self.delta = delta
        self.species = {}
        self.genomes = []
        self.generation = 0
        self.metric = metric
        genome_seeder = genome_seeder(population_size) \
            if isgeneratorfunction(genome_seeder) else genome_seeder
        self.genomes = [genome for genome in genome_seeder]

    def speciate(self):
        """Divides the population of genomes up into species.

        :return: None
        """
        self.species[1] = {
            'repr': self.genomes[0],
            'group': [self.genomes[0]]
        }

        for genome in self.genomes[1:]:
            assigned_group = False
            for key, item in self.species.items():
                if self.delta is None or self.metric is None or \
                        self.metric(genome, item['repr']) < self.delta:
                    assigned_group = True
                    self.species[key]['group'].append(genome)
                    break
            if not assigned_group:
                self.species[len(self.species) + 1] = {
                    'repr': genome,
                    'group': [genome]
                }

        for key, item in self.species.items():
            group_size = len(item['group'])
            adj_fitness = lambda x: x.fitness / group_size
            group_fitness = sum([adj_fitness(g) for g in item['group']])
            item['group_fitness'] = group_fitness
            item['group'].sort(key=adj_fitness, reverse=True)
            item['generation'] = self.generation

    def to_dict(self):
        """Returns dictionary summary of population.

        :return: dictionary summary of population.
        """
        best_genome = max(self.genomes, key=lambda g: g.fitness)
        worst_genome = min(self.genomes, key=lambda g: g.fitness)
        data = {
            'generation': self.generation,
            'species_num': len(self.species),
            'best_fitness': best_genome.fitness,
            'best_genome': best_genome.to_reduced_repr,
            'worst_fitness': worst_genome.fitness,
            'worst_genome': worst_genome.to_reduced_repr,
            'mean_fitness': np.mean([g.fitness for g in self.genomes]),
            'groups': [],
            'created_at': time()
        }
        if self.metric is not None:
            for key, item in self.species.items():
                if item['generation'] != self.generation:
                    raise ValueError('current generation and species group generation are not the same!')
                data['groups'].append({
                    'id': key,
                    'bests': [g.to_reduced_repr for g in item['group'][0:3]],
                    'group_fitness': item['group_fitness'],
                    'repr': item['repr'].to_reduced_repr,
                    'partition': {
                        key: self.metric(item['repr'], other_item['repr'])
                        for key, other_item in self.species.items()
                    }
                })
        return data



import numpy as np


def vector_decomp(p, o, v):
    """
    returns the parallel and orthogonal distances of the point
    along the vector v from o. i.e.

    dist(p, l(t) = o + tv) = min_{t}(dist(p, o + tv))


    :param p: point to decompose into v_orthogonal and v_parallel
    :param o: origin of line.
    :param v: vector along which to decompose p
    :return: p_parallel, p_orthogonal
    """

    v = np.array(v) / np.linalg.norm(v)
    p, o = (np.array(p), np.array(o))
    p_parallel = np.dot(o - p, v)
    p_orthogonal = np.linalg.norm((o - p) - p_parallel*v)
    return -p_parallel, p_orthogonal



"""Class that acts on a genome or population of genomes.

The call_on_population and call_on_genome methods are
expected to be overwritten.
"""

from gerel.genome.genome import Genome
from gerel.populations.population import Population


class Mutator:
    def __init__(self):
        pass

    def __call__(self, target):
        """Mutates a target genome or population in place.

        :param target: Population or Genome.
        :return: None
        :raises ValueError: If target not a Genome or Population.
        """
        if isinstance(target, Genome):
            self.call_on_genome(target)
        elif isinstance(target, Population):
            self.call_on_population(target)
        else:
            raise ValueError('Target should be either a Genome or a Population')

    def call_on_population(self, population):
        err_msg = 'Mutator call should act on Population Objects. Did you forget to overwrite call_on_population'
        raise NotImplementedError(err_msg)

    def call_on_genome(self, genome):
        err_msg = 'Mutator call should act on Genome Objects. Did you forget to overwrite call_on_genome'
        raise NotImplementedError(err_msg)



from gerel.algorithms.NEAT.metric import generate_neat_metric
from gerel.debug.class_debug_decorator import add_inst_validator
from gerel.debug.population_validator import validate_population
from gerel.populations.population import Population
from gerel.populations.genome_seeders import curry_genome_seeder
from gerel.algorithms.NEAT.mutator import NEATMutator


@add_inst_validator(env="TESTING", validator=validate_population)
class NEATPopulation(Population):
    def __init__(
            self,
            population_size=150,
            delta=3.0,
            genome_seeder=curry_genome_seeder(NEATMutator()),
            metric=generate_neat_metric(1, 1, 3)):
        """Speciated Population of genomes used within the NEAT algorithm. Extends Population class.

        Notes:
          - The distance metric allows us to speciate using a compatibility threshold
          delta. An ordered list of species is maintained. In each generation, genomes are
          sequentially placed into species.

        :param genome_seeder: An iterable of genomes that will be used to generate the
            initial population.
        :param population_size: An integer number giving the size of the total Population.
        :param delta: A float number that defines the minimum distance two genomes require
            to be apart before there sorted into different species. Set to None if
            single species is desired (no speciation).
        :param metric: Function that takes two genomes and returns the distance between them.
            The metric is used to partition the genomes into species. If None then no
            speciation will take place.
        """
        super().__init__(population_size=population_size,
                         delta=delta,
                         genome_seeder=genome_seeder,
                         metric=metric)



import numpy as np
from math import exp


def step(x):
    return 1 if x > 0 else - 1


def relu(x):
    return max(0, x)


def tanh(x):
    return np.tanh(x)


def build_sigmoid(c=4.9):
    def sigmoid(x):
        exp_arg = c * x
        if - exp_arg > 100:
            return 0
        return 1 / (1 + exp(-exp_arg))
    return sigmoid


def build_leaky_relu(c=0.01):
    def leaky_relu(x):
        return c * x if x < 0 else x
    return leaky_relu



"""
Simple wrapper to compute a parrallelize a function.

given a function f that takes a single argument, `arg`, the batch_job decorator
turns f into a function that takes a list of args and computes each result in
a seperate process. It then returns the results as a list.

___

use:
    ```
    batch_job = BatchJob(num_processes=3)

    @batch_job
    def job(arg):
        # do something...

    job([args_1, args_2, ....])
    ```

The above runs: job(args_1), job(args_2), ... in parrallel and returns
[job(args_1), job(args_2), ...]
"""


from multiprocessing import Process, Queue, JoinableQueue, cpu_count

PROCESSES = cpu_count() - 1


class BatchJob:
    def __init__(self, num_processes=PROCESSES):
        self.in_queue = JoinableQueue()
        self.out_queue = Queue()
        self.num_processes = num_processes

    def __call__(self, job):
        def batched_job(arg_batch):
            def process_task(task_queue, return_queue):
                while True:
                    next_task = task_queue.get()
                    if next_task is None:
                        task_queue.task_done()
                        break
                    args_ind, args = next_task
                    result = job(args)
                    task_queue.task_done()
                    return_queue.put((args_ind, result))
                return True

            for i in range(self.num_processes):
                p = Process(target=process_task,
                            args=(self.in_queue, self.out_queue))
                p.start()

            for ind, arg in enumerate(arg_batch):
                self.in_queue.put((ind, arg))

            for _ in range(self.num_processes):
                self.in_queue.put(None)

            self.in_queue.join()

            results = []
            while not self.out_queue.empty():
                results.append(self.out_queue.get())
            results.sort(key=lambda item: item[0])
            return [r for i, r in results]
        return batched_job


def partition(ls, ls_size):
    """Partitions list into a list of lists.

    Useful for dividing up a population into batches in order to
    process each batch in a different process.

    :param ls: list to be partitioned.
    :param ls_size: length of sublist.
    :return: partitioned list
    """
    parition_num = int(len(ls) / ls_size) + 1
    return [ls[i * ls_size:(i + 1) * ls_size] for i in range(parition_num)]


def departition(ls):
    """Reverse of partition.

    Takes list of lists and turn it into a list.

    :param ls: List to be departitioned
    :return: departitioned list
    """
    return [item for sublist in ls for item in sublist]



"""Factory methods for building specific genome architectures or copying genomes."""

from gerel.genome.genome import Genome
from gerel.genome.node import Node
from gerel.genome.edge import Edge


def from_genes(
        nodes_genes,
        edges,
        input_size=2,
        output_size=2,
        weight_low=-2,
        weight_high=2,
        depth=3):
    """Reconstructs a Genome class from the return value
    to_reduced_repr called on an instance of a Genome class.

    :param nodes_genes: First member of genome.to_reduced_repr,
        list of node properties and indices tuples. Takes form:

        [(layer_num, layer_ind, innov, weight, type), ...]

        where layer number is the genome layer index, layer_ind
        is the index in the layer, innov is the innovation number.
        weight is the node bias and type is one of 'input',
        'hidden' or 'output'.
    :param edges: Second member of genome.to_reduced_repr,
        tuple of edge properties and indices. Takes form:

        [(
            from_node.to_reduced_repr,
            to_node.to_reduced_repr,
            weight,
            innov,
            active
        ), ...]

        where from_node.to_reduced_repr and
        to_node.to_reduced_repr take the same form as an
        individual node in node_genes. weight is the
        edge weight, innov it's innovation number and
        active is True or False.
    :param input_size: Number of input nodes
    :param output_size: Number of output nodes
    :param weight_low: Maximum weight on node and edges
    :param weight_high: Minimum weight on node and edges
    :param depth: Number of layers in network.
    :return: Constructed genome.
    """

    new_genome = Genome(
        input_size=input_size,
        output_size=output_size,
        weight_low=weight_low,
        weight_high=weight_high,
        depth=depth)

    layer_maxes = [0 for _ in range(depth)]
    for node_gene in nodes_genes:
        layer_num, layer_ind, _, weight, _ = node_gene
        if layer_maxes[layer_num - 1] < layer_ind + 1:
            layer_maxes[layer_num - 1] = layer_ind + 1

    layers = []
    for layer_max in layer_maxes:
        layers.append([None for _ in range(layer_max)])

    nodes = []
    for node_gene in nodes_genes:
        layer_num, layer_ind, innov, weight, node_type = node_gene
        node = Node(layer_num, layer_ind, weight, type=node_type)
        nodes.append(node)
        layers[layer_num - 1][layer_ind] = node

    new_genome.layers = [new_genome.inputs, *layers, new_genome.outputs]

    new_genome.nodes = nodes
    for from_node_reduced, to_node_reduced, weight, innov, active in edges:
        from_layer_num, from_layer_ind, _, _, _ = from_node_reduced
        from_node = new_genome.layers[from_layer_num][from_layer_ind]
        to_layer_num, to_layer_ind, _, _, _ = to_node_reduced
        to_node = new_genome.layers[to_layer_num][to_layer_ind]
        edge = Edge(from_node, to_node, weight, active)
        new_genome.edges.append(edge)
        new_genome.edge_innovs.add((from_node.innov, to_node.innov))
    return new_genome


def minimal(
        input_size=2,
        output_size=2,
        weight_low=-2,
        weight_high=2,
        depth=3):
    """ Builds a minimal genome with specified inputs and
    outputs, weight bounds, depth and one connected node in
    the first layer.

    :param input_size: Number of input nodes
    :param output_size: Number of output nodes
    :param weight_low: Maximum weight on node and edges
    :param weight_high: Minimum weight on node and edges
    :param depth: Number of layers in network.
    :return: Constructed genome.
    """

    genome = Genome(
        input_size=input_size,
        output_size=output_size,
        weight_low=weight_low,
        weight_high=weight_high,
        depth=depth)
    genome.layers = [genome.inputs,
                     *[[] for _ in range(depth)],
                     genome.outputs]
    genome.add_node(1)
    for n in genome.inputs:
        genome.add_edge(n, genome.layers[1][0])
    for n in genome.outputs:
        genome.add_edge(genome.layers[1][0], n)
    return genome


def dense(
        input_size=2,
        output_size=2,
        weight_low=-2,
        weight_high=2,
        layer_dims=(5, 5, 5)):
    """ Builds a densely connected genome with specified
    inputs and outputs, weight bounds and depth with every
    node in one layer connected to every node in the
    subsequent layer.


    :param input_size: Number of input nodes
    :param output_size: Number of output nodes
    :param weight_low: Maximum weight on node and edges
    :param weight_high: Minimum weight on node and edges
    :param layer_dims: Tuple of layer dimensions
    :return: Constructed genome.
    """

    genome = Genome(
        input_size=input_size,
        output_size=output_size,
        weight_low=weight_low,
        weight_high=weight_high,
        depth=len(layer_dims))

    genome.layers = [genome.inputs,
                     *[[] for _ in range(len(layer_dims))],
                     genome.outputs]

    layer_ind = 1
    for layer_size in layer_dims:
        for _ in range(layer_size):
            genome.add_node(layer_ind)
        layer_ind += 1

    for layer_1, layer_2 in zip(genome.layers, genome.layers[1:]):
        for node_1 in layer_1:
            for node_2 in layer_2:
                genome.add_edge(node_1, node_2)

    return genome


def copy(genome):
    """Deep copy of genome instance.

    :param genome: Instance of Genome class
    :return: Copied instance of Genome glass
    """
    new_genome = Genome(
        input_size=len(genome.inputs),
        output_size=len(genome.outputs),
        weight_low=genome.weight_low,
        weight_high=genome.weight_high,
        depth=genome.depth)
    layers = [[Node.copy(node) for node in layer]
              for layer in genome.layers[1:-1]]
    new_genome.layers = [
        new_genome.inputs,
        *layers,
        new_genome.outputs
    ]
    nodes = [new_genome.layers[node.layer_num][node.layer_ind]
             for node in genome.nodes]
    new_genome.nodes = nodes
    for edge in genome.edges:
        new_edge = Edge.copy(edge, new_genome)
        new_genome.edge_innovs.add((
            new_edge.from_node.innov,
            new_edge.to_node.innov))
    return new_genome



"""Compatibility Distance:

    δ = c_1*E/N + c_2*D/N + c3 · W.

where E is excess and D is disjoint genes and W is the average weight
differences of matching genes including disabled genes.

From NEAT paper http://nn.cs.utexas.edu/downloads/papers/stanley.ec02.pdf
"""


def generate_gene_metric(c_1=1.0, c_2=1.0, c_3=0.4):
    def _compare_gene_difference(genes_1, genes_2):
        last_match = [0, 0]
        W, M, i, j = (0, 0, 0, 0)
        while True:
            if genes_1[i].innov == genes_2[j].innov:
                last_match = [i, j]
                W += abs(genes_1[i].weight - genes_2[j].weight)
                M, i, j = (M + 1, i + 1, j + 1)
            elif genes_1[i].innov < genes_2[j].innov:
                i += 1
            elif genes_1[i].innov > genes_2[j].innov:
                j += 1
            if i == len(genes_1) or j == len(genes_2):
                break
        N = max(len(genes_1), len(genes_2))
        D = 2 + last_match[0] + last_match[1] - 2 * M
        E = len(genes_1) + len(genes_2) - last_match[0] - last_match[1] - 2
        W = W / M
        return c_1 * E / N + c_2 * D / N + c_3 * W
    return _compare_gene_difference


def generate_neat_metric(c_1=1.0, c_2=1.0, c_3=0.4):
    """Builds a metric on reduced representation of genomes."""
    gen_metric = generate_gene_metric(c_1, c_2, c_3)

    def metric(genome_1, genome_2):
        d = gen_metric(genome_1.nodes, genome_2.nodes)
        d += gen_metric(genome_1.edges, genome_2.edges)
        return d
    return metric



def validate_genome(genome, *args, **kwargs):
    if len(set(e.innov for e in genome.edges)) != len(genome.edges):
        raise ValueError('Non-unique edge in genome edges.')

    if len(set(n.innov for n in genome.nodes)) != len(genome.nodes):
        raise ValueError('Non-unique node in genome node.')

    for e in genome.edges:
        if not (e.from_node.innov, e.to_node.innov) in genome.edge_innovs:
            raise ValueError('Edge innovation not registered.')

    for e_1, e_2 in zip(genome.edges, genome.edges[1:]):
        if not e_1.innov < e_2.innov:
            raise ValueError('Edge innovations are out of order.')

    for n_1, n_2 in zip(genome.nodes, genome.nodes[1:]):
        if not n_1.innov < n_2.innov:
            raise ValueError('Node innovations are out of order.')

    for n in genome.inputs:
        if not n.type == 'input':
            raise ValueError('Non input node in inputs.')

    for n in genome.outputs:
        if not n.type == 'output':
            raise ValueError('Non output node in outputs.')



import numpy as np
from numpy.random import choice, uniform
from gerel.genome.factories import from_genes


def curry_weight_mutator(
        weight_mutation_likelihood,
        weight_mutation_rate_random,
        weight_mutation_variance,
        weight_mutation_rate_uniform,
        weight_low=-2,
        weight_high=2,
        ):
    """Creates function that acts on nodes or edges to mutate
    them.

    :param weight_mutation_likelihood: Given a node or edge this
        is the likelihood that targets weight is mutated.
    :param weight_mutation_rate_random: Likelihood of normal
        distribution perturbation of weight.
    :param weight_mutation_variance: Variance of normal distribution
        used to perturb weights.
    :param weight_mutation_rate_uniform: Likelihood of uniform
        distribution perturbation of weight.
    :param weight_low: uniform distribution lower bound
    :param weight_high: uniform distribution upper bound
    :return: Function that acts on Nodes or Edges to mutate them.
    """
    def weight_adj(target):
        if np.random.uniform(0, 1, 1) < weight_mutation_likelihood:
            if np.random.uniform(0, 1, 1) < weight_mutation_rate_random:
                perturbation = np.random \
                    .normal(0, weight_mutation_variance, 1)[0]
                target.weight += perturbation
            if weight_mutation_rate_random < np.random.uniform(0, 1) < \
                    weight_mutation_rate_random + weight_mutation_rate_uniform:
                perturbation = np.random.uniform(
                    weight_low,
                    weight_high)
                target.weight = perturbation
    return weight_adj


def pair_genes(primary, secondary):
    """Crossover function

    For any list of genes with innov values (edges or nodes) we
    iterate through and chose one randomly when there exists a
    corresponding innov number and select the primary (fittest) gene when
    a match doesn't exist.

    :param primary: Edges or nodes belonging to genome with
        greater fitness than secondary.
    :param secondary: Edges or nodes belonging to genome with
        lesser fitness than primary.
    :return: List of selected reduced representation edges or
        nodes.
    """

    i, j = (0, 0)
    selected_gene = []
    while True:
        if primary[i].innov == secondary[j].innov:
            gene = choice([primary[i], secondary[j]])
            selected_gene.append(gene.to_reduced_repr)
            i, j = (i + 1, j + 1)
        elif primary[i].innov < secondary[j].innov:
            selected_gene.append(primary[i].to_reduced_repr)
            i += 1
        elif primary[i].innov > secondary[j].innov:
            j += 1

        if i == len(primary):
            break

        if j == len(secondary):
            excess = [gene.to_reduced_repr for gene in primary[i:]]
            selected_gene = selected_gene + excess
            break
    return selected_gene


def add_edge(genome):
    """Random edge mutation on genome.

    When a edge is added we sample two layers without replacement and
    then order them. We then sample a node from each and add a new edge
    between them.

    :param genome: Genome being mutated.
    :return: Genome being mutated.
    """

    non_empty_layers = [layer_num for layer_num in
                        range(len(genome.layers)) if
                        len(genome.layers[layer_num]) > 0]
    from_layer, to_layer = sorted(choice(
        non_empty_layers, 2, replace=False))
    from_node = choice(genome.layers[from_layer])
    to_node = choice(genome.layers[to_layer])
    genome.add_edge(from_node, to_node)
    return genome


def add_node(genome):
    """Random Node mutation.

    When a node is added we randomly sample an admissible edge and then
    randomly sample an layer index in the range of layers the edge spans.
    After disabling the sampled edge we add a new node in the selected
    layer and then connect it with two new edges.

    :param genome: Genome being mutated.
    :return: Genome being mutated.
    """

    edge = choice(genome.get_admissible_edges())
    from_node, to_node = (edge.from_node, edge.to_node)
    layer_num = choice(range(from_node.layer_num + 1, to_node.layer_num))
    new_node = genome.add_node(layer_num)
    genome.add_edge(from_node, new_node)
    genome.add_edge(new_node, to_node)
    edge.active = False
    return genome


def curry_crossover(gene_disable_rate):
    """Creates a crossover function.

    Creates a crossover function that takes a more fit and less
    fit genome, pairs there edge and node genes and selects one
    from each at random. If a disabled edge is selected it's
    reactivated with 1 - gene_disable_rate probability.

    :param gene_disable_rate: Likelihood of disabled gene staying
        inactive.
    :return: New genome with parents primary and secondary.
    """
    def crossover(primary=None, secondary=None):
        """Produces a child of the primary and secondary genomes.

        Note that genome.nodes excludes input and output layers.
        """
        node_genes = pair_genes(
            primary.nodes,
            secondary.nodes)
        edge_genes = pair_genes(
            primary.edges,
            secondary.edges)

        def activate_disabled(edge):
            (a, b, c, d, active) = edge
            if active:
                return edge
            if uniform(0, 1) > gene_disable_rate:
                edge = (a, b, c, d, True)
            return edge

        edge_genes = [activate_disabled(edge) for edge in edge_genes]
        new_genome = from_genes(
            node_genes,
            edge_genes,
            input_size=len(primary.inputs),
            output_size=len(primary.outputs),
            weight_low=primary.weight_low,
            weight_high=primary.weight_high,
            depth=primary.depth)
        return new_genome
    return crossover



from gerel.debug.class_debug_decorator import add_inst_validator
from gerel.debug.population_validator import validate_population
from gerel.populations.population import Population


@add_inst_validator(env="TESTING", validator=validate_population)
class SIMPLEPopulation(Population):
    def __init__(
            self,
            genome_seeder,
            population_size=150):
        """Non-Speciated Population of genomes used within the SIMPLE algorithm.

        Extends Population class.

        :param genome_seeder: An iterable of genomes that will
            be used to generate the initial population.
        :param population_size: An integer number giving the size
            of the total Population.
        """
        super().__init__(population_size=population_size,
                         delta=None,
                         genome_seeder=genome_seeder,
                         metric=None)



from gerel.debug.class_debug_decorator import add_inst_validator
from gerel.debug.population_validator import validate_population
from gerel.populations.population import Population


@add_inst_validator(env="TESTING", validator=validate_population)
class RESPopulation(Population):
    def __init__(
            self,
            genome_seeder,
            population_size=150):
        """Population of genomes used within the RES algorithm.

        :param genome_seeder: An iterable of genomes that will
            be used to generate the initial population.
        :param population_size: An integer number giving the size
            of the total Population.
        """
        # if genome_seeder
        super().__init__(population_size=population_size,
                         delta=None,
                         genome_seeder=genome_seeder,
                         metric=None)

    def rank_transform(self):
        """Orders targets by there fitness and then allocates each an
        evenly distributed fitness between -0.5, 0.5.
        """
        sorted_targets = sorted(self.genomes, key=lambda item: item.fitness)
        for index, target in enumerate(sorted_targets):
            target.fitness = index/len(sorted_targets) - 0.5



"""Functions that can be used as decorators to check the state of class datastructures are sound.

The class_validator function adds a validation step to each method in the decorated class if the environment is
in TESTING mode. This is used primarily for testing in order to catch cases where class instances enter states
they shouldn't.

Example:

```
def simple_validator(self, *args, **kwargs):
    if self.value > 10:
        raise ValueError()


@add_inst_validator(env="TESTING", validator=simple_validator)
class Test:
    def __init__(self, value):
        self.value = value

    def set_value(self, value):
        self.value = value
```

"""
import functools


def print_all(cls_inst, *args, **kwargs):
    print('instance:', cls_inst)
    print('args:', *args)
    print('kwargs:', **kwargs)


def decorator(func, validator):
    @functools.wraps(func)
    def new_method(*args, **kwargs):
        ret_val = func(*args, **kwargs)
        validator(*args, **kwargs)
        return ret_val
    return new_method


def add_inst_validator(env='TESTING', validator=print_all):
    def class_decorator(cls):
        if env == 'TESTING':
            for attr, val in cls.__dict__.items():
                if callable(val) and not attr.startswith("__"):
                    setattr(cls, attr, decorator(val, validator))
        return cls
    return class_decorator
